/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_socks.c
 *   Copyright (C) 1998        Lucas Madar
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "patchlevel.h"
#include "nameser.h"
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#if defined(__hpux)
# include "inet.h"
#endif
#include <fcntl.h>
#include "sock.h"
#include "h.h"

CVSID("$Id: socks.c,v 1.10 2001/01/16 03:48:38 smkelly Exp $");

static unsigned char socksid[12];

#define HICHAR(s)      (((unsigned short) s) >> 8)
#define LOCHAR(s)      (((unsigned short) s) & 0xFF)

#ifndef SOCKSPORT
# define SOCKSPORT 6013
#endif

/* init_socks
 * Set up socksid, simply.
 *
 * IPv6: The initialization can go on... but the check can only be made against IPv4 addresses...
 *       Even though we allocate an IPv6 socket <grin>
 *       NOTE: On OS'es like NT where the IPv4 stack is completely seperated from the IPv6 stack
 *             and vice versa this trick doesn't work... one needs to bind to different sockets.
 *             being.. one for IPv4 and one for IPv6...
 */
void init_socks(aClient *cptr)
{
	unsigned short sport = SOCKSPORT;
	struct SOCKADDR sin;

	socksid[0] = 4;
	socksid[1] = 1;
	socksid[2] = HICHAR(sport);
	socksid[3] = LOCHAR(sport);
	socksid[8] = 0;

	if ((cptr->socksfd = socket(AFINET, SOCK_STREAM, 0)) == -1)
		return;
	
	set_non_blocking(cptr->socksfd, cptr);
#ifdef IPV6
	bcopy(me.ip.s6_addr, cptr->ip.s6_addr, IN6ADDRSZ);
#else
	cptr->ip.s_addr = me.ip.S_ADDR
#endif
	sin.SIN_PORT = htons(sport);
	sin.SIN_FAMILY = AFINET;

	if (bind(cptr->socksfd, (struct sockaddr *)&sin, sizeof(sin)))
	{
		close(cptr->socksfd);
		cptr->socksfd = -1;
		return;
	}

	listen(cptr->socksfd, LISTEN_SIZE);
	
	/* Socks listening port is now set up */
}

/*
 *
 * attempts to connect to a socks server on port 1080 of the host.
 * if the connect fails, the user passes the socks check
 * otherwise, the connection is checked to see if it is a "secure"
 * socks4+ server
 */
 void	start_socks(cptr)
 aClient	*cptr;
 {
 	struct SOCKADDR_IN sin;
	int sinlen = sizeof(struct SOCKADDR_IN);
#ifdef IPV6
	u_char *cp = (u_char *)cptr->ip.s6_addr;
#endif
	if (!S_SOCKSCHECK) return;
#ifdef IPV6
	/* Verify if this is actually IPv4, otherwise we won't
		do a SOCKS check as SOCKS v4 simply doesn't support IPv6 addressing anyway :)
		A seperate SOCKS v5 checker will be added to also check IPv6 addresses... -Fuzzel */
	// Is it actually an IPv4 address??
	if (!( cp[ 0] == 0 && cp[ 1] == 0  &&  cp[ 2] == 0    && cp[ 3] == 0    && cp[4] == 0 &&
			 cp[ 5] == 0 && cp[ 6] == 0  &&  cp[ 7] == 0    && cp[ 8] == 0    && cp[9] == 0 &&
			((cp[10] == 0 && cp[11] == 0) || (cp[10] == 0xff && cp[11] == 0xff))))
	{
		// It is a non-IPv4/real-IPv6 address :(
		// Set the socksfd to -1 just to be sure...
		cptr->socksfd = -1;
		Debug((DEBUG_INFO, "Skipping IPv4 Socks Check..."));
		return;
	}
#endif
	Debug((DEBUG_INFO, "Doing IPv4 Socks Check..."));
	
	if ((cptr->socksfd = socket(AFINET, SOCK_STREAM, 0)) == -1)
	{
		Debug((DEBUG_ERROR, "Unable to create socks socket for %s:%s",
			get_client_name(cptr, TRUE),
			strerror(get_sockerr(cptr))));
		return;
	}
	if (cptr->socksfd >= (MAXCONNECTIONS - 3))
	{
		sendto_ops("Can't allocate fd for socks on %s",
			get_client_name(cptr, TRUE));
		close(cptr->socksfd);
		cptr->socksfd = -1;
		return;
	}

	if (SHOWCONNECTINFO)
		write(cptr->fd, REPORT_DO_SOCKS, R_do_socks);

	set_non_blocking(cptr->socksfd, cptr);

	sin.SIN_PORT = htons(1080);
	sin.SIN_FAMILY = AFINET;
	bcopy((char *)&cptr->ip, (char *)&sin.SIN_ADDR, sizeof(struct IN_ADDR));
	
	if (connect(cptr->socksfd, (struct sockaddr *)&sin, sinlen) == -1 && errno != EINPROGRESS)
	{
		/* we have no socks server! */
		close(cptr->socksfd);
		cptr->socksfd = -1;
		if (SHOWCONNECTINFO)
			write(cptr->fd, REPORT_NO_SOCKS, R_no_socks);
		return;
	}
	cptr->flags |= (FLAGS_WRSOCKS|FLAGS_SOCKS);
	if (cptr->socksfd > highest_fd)
		highest_fd = cptr->socksfd;
	return;
}

/*
 *
 * send the socks server a query to see if it's open.
 */
void	send_socksquery(cptr)
aClient	*cptr;
{
	struct SOCKADDR sin;
	int sinlen = sizeof(struct SOCKADDR_IN);
	unsigned char socksbuf[12];
	unsigned long theip;

	bcopy((char *)&socksid, (char *)&socksbuf, 9);
	getsockname(cptr->fd, (struct sockaddr *)&sin, &sinlen);
#ifdef IPV6
	{
		u_char *cp = (u_char *)sin.sin6_addr.s6_addr;
		if ( cp[ 0] == 0 && cp[ 1] == 0  &&  cp[ 2] == 0    && cp[ 3] == 0    && cp[4] == 0 &&
				 cp[ 5] == 0 && cp[ 6] == 0  &&  cp[ 7] == 0    && cp[ 8] == 0    && cp[9] == 0 &&
				((cp[10] == 0 && cp[11] == 0) || (cp[10] == 0xff && cp[11] == 0xff)))
		{
			/* It's IPv4... so copy and check it... */
			socksbuf[4] = cp[12];
			socksbuf[5] = cp[13];
			socksbuf[6] = cp[14];
			socksbuf[7] = cp[15];
		}
		else
		{
			/* Grr... and how did that come until here? (Should not happen as it gets exited by start_socks)
				It's IPv6... can't check that.. won't even fit into the SOCKS struct :) */
			return;
		}
	}
#else
	theip = htonl(sin.SIN_ADDR.S_ADDR);
	socksbuf[4] = (theip >> 24);
	socksbuf[5] = (theip >> 16) & 0xFF;
	socksbuf[6] = (theip >> 8) & 0xFF;
	socksbuf[7] = theip & 0xFF;
#endif

	if (send(cptr->socksfd, socksbuf, 9, 0) != 9)
	{
		close(cptr->socksfd);
		if (cptr->socksfd == highest_fd)
			while (!local[highest_fd])
				highest_fd--;
		cptr->socksfd = -1;
		cptr->flags &= ~FLAGS_SOCKS;
		if (SHOWCONNECTINFO)
			write(cptr->fd, REPORT_NO_SOCKS, R_no_socks);
	}
	cptr->flags &= ~FLAGS_WRSOCKS;
	return;
}

/*
 * read_socks
 *
 * process the socks reply.
 */
void	read_socks(cptr)
aClient	*cptr;
{
	unsigned char socksbuf[12];
	int len;

	len = recv(cptr->socksfd, socksbuf, 9, 0);

	(void)close(cptr->socksfd);
	if (cptr->socksfd == highest_fd)
		while (!local[highest_fd])
			highest_fd--;
	cptr->socksfd = -1;
	ClearSocks(cptr);

	if(len < 2)
	{
		if (SHOWCONNECTINFO)
			write(cptr->fd, REPORT_NO_SOCKS, R_no_socks);
		return;
	}

	if (socksbuf[1] == 90)
	{
		cptr->flags |= FLAGS_GOTSOCKS;
		return;
	}
	if (SHOWCONNECTINFO)
		write(cptr->fd, REPORT_GOOD_SOCKS, R_good_socks);
	return;
}


/*
 * $Log: socks.c,v $
 * Revision 1.10  2001/01/16 03:48:38  smkelly
 * Removed all WIN32 code
 *
 * Revision 1.9  2000/12/13 10:16:14  smkelly
 * Removed some _WIN32 code
 *
 * Revision 1.8  2000/07/19 04:35:46  smkelly
 * SET SHOWALL and SET CONNINFO
 *
 * Revision 1.7  2000/07/19 04:03:33  smkelly
 * SET SOCKSCHECK
 *
 * Revision 1.6  2000/07/19 03:35:54  smkelly
 * UNIXPORT support removed
 *
 * Revision 1.5  1999/11/13 09:54:49  smkelly
 * SOCKS check corrected for SOCKS4 courtesy of Lucas Madar
 *
 * Revision 1.4  1999/11/12 11:12:10  smkelly
 * Added RCS tags
 *
 */
 
